'use strict';

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _graphql = require('graphql');

var _lodash = require('lodash');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var graphQLValidationRuleNames = ['UniqueOperationNames', 'LoneAnonymousOperation', 'KnownTypeNames', 'FragmentsOnCompositeTypes', 'VariablesAreInputTypes', 'ScalarLeafs', 'FieldsOnCorrectType', 'UniqueFragmentNames',
//'KnownFragmentNames', -> any interpolation
//'NoUnusedFragments', -> any standalone fragment
'PossibleFragmentSpreads', 'NoFragmentCycles', 'UniqueVariableNames', 'NoUndefinedVariables', 'NoUnusedVariables', 'KnownDirectives', 'KnownArgumentNames', 'UniqueArgumentNames', 'ArgumentsOfCorrectType', 'ProvidedNonNullArguments', 'DefaultValuesOfCorrectType', 'VariablesInAllowedPosition', 'OverlappingFieldsCanBeMerged', 'UniqueInputFieldNames'];

// Omit these rules when in Relay env
var relayRuleNames = (0, _lodash.without)(graphQLValidationRuleNames, 'ScalarLeafs', 'ProvidedNonNullArguments', 'KnownDirectives', 'NoUndefinedVariables');

var graphQLValidationRules = graphQLValidationRuleNames.map(function (ruleName) {
  return require('graphql/validation/rules/' + ruleName)[ruleName];
});

var relayGraphQLValidationRules = relayRuleNames.map(function (ruleName) {
  return require('graphql/validation/rules/' + ruleName)[ruleName];
});

var internalTag = 'ESLintPluginGraphQLFile';
var gqlFiles = ['gql', 'graphql'];

var rules = {
  'template-strings': {
    meta: {
      schema: {
        type: 'array',
        minLength: 1,
        items: {
          additionalProperties: false,
          properties: {
            schemaJson: {
              type: 'object'
            },
            schemaJsonFilepath: {
              type: 'string'
            },
            env: {
              enum: ['lokka', 'relay', 'apollo']
            },
            tagName: {
              type: 'string',
              pattern: '^[$_a-zA-Z$_][a-zA-Z0-9$_]+(\\.[a-zA-Z0-9$_]+)?$'
            }
          },
          // schemaJson and schemaJsonFilepath are mutually exclusive:
          oneOf: [{
            required: ['schemaJson'],
            not: { required: ['schemaJsonFilepath'] }
          }, {
            required: ['schemaJsonFilepath'],
            not: { required: ['schemaJson'] }
          }]
        }
      }
    },
    create: function create(context) {
      var tagNames = new Set();
      var tagRules = [];
      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = context.options[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          var optionGroup = _step.value;

          var _parseOptions = parseOptions(optionGroup),
              schema = _parseOptions.schema,
              env = _parseOptions.env,
              tagName = _parseOptions.tagName;

          if (tagNames.has(tagName)) {
            throw new Error('Multiple options for GraphQL tag ' + tagName);
          }
          tagNames.add(tagName);
          tagRules.push({ schema: schema, env: env, tagName: tagName });
        }
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return) {
            _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }

      return {
        TaggedTemplateExpression: function TaggedTemplateExpression(node) {
          var _iteratorNormalCompletion2 = true;
          var _didIteratorError2 = false;
          var _iteratorError2 = undefined;

          try {
            for (var _iterator2 = tagRules[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
              var _ref2 = _step2.value;
              var schema = _ref2.schema,
                  env = _ref2.env,
                  tagName = _ref2.tagName;

              if (templateExpressionMatchesTag(tagName, node)) {
                return handleTemplateTag(node, context, schema, env);
              }
            }
          } catch (err) {
            _didIteratorError2 = true;
            _iteratorError2 = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion2 && _iterator2.return) {
                _iterator2.return();
              }
            } finally {
              if (_didIteratorError2) {
                throw _iteratorError2;
              }
            }
          }
        }
      };
    }
  }
};

function parseOptions(optionGroup) {
  var schemaJson = optionGroup.schemaJson,
      schemaJsonFilepath = optionGroup.schemaJsonFilepath,
      env = optionGroup.env,
      tagNameOption = optionGroup.tagName;

  // Validate and unpack schema

  var schema = void 0;
  if (schemaJson) {
    schema = initSchema(schemaJson);
  } else if (schemaJsonFilepath) {
    schema = initSchemaFromFile(schemaJsonFilepath);
  } else {
    throw new Error('Must pass in `schemaJson` option with schema object ' + 'or `schemaJsonFilepath` with absolute path to the json file.');
  }

  // Validate env
  if (env && env !== 'lokka' && env !== 'relay' && env !== 'apollo') {
    throw new Error('Invalid option for env, only `apollo`, `lokka`, and `relay` supported.');
  }

  // Validate tagName and set default
  var tagName = void 0;
  if (tagNameOption) {
    tagName = tagNameOption;
  } else if (env === 'relay') {
    tagName = 'Relay.QL';
  } else {
    tagName = 'gql';
  }
  return { schema: schema, env: env, tagName: tagName };
}

function initSchema(json) {
  var unpackedSchemaJson = json.data ? json.data : json;
  if (!unpackedSchemaJson.__schema) {
    throw new Error('Please pass a valid GraphQL introspection query result.');
  }
  return (0, _graphql.buildClientSchema)(unpackedSchemaJson);
}

function initSchemaFromFile(jsonFile) {
  return initSchema(JSON.parse(_fs2.default.readFileSync(jsonFile, 'utf8')));
}

function templateExpressionMatchesTag(tagName, node) {
  var tagNameSegments = tagName.split('.').length;
  if (tagNameSegments === 1) {
    // Check for single identifier, like 'gql'
    if (node.tag.type !== 'Identifier' || node.tag.name !== tagName) {
      return false;
    }
  } else if (tagNameSegments === 2) {
    // Check for dotted identifier, like 'Relay.QL'
    if (node.tag.type !== 'MemberExpression' || node.tag.object.name + '.' + node.tag.property.name !== tagName) {
      return false;
    }
  } else {
    // We don't currently support 3 segments so ignore
    return false;
  }
  return true;
}

function handleTemplateTag(node, context, schema, env) {
  var text = void 0;
  try {
    text = replaceExpressions(node.quasi, context, env);
  } catch (e) {
    if (e.message !== 'Invalid interpolation') {
      console.log(e);
    }
    return;
  }

  // Re-implement syntax sugar for fragment names, which is technically not valid
  // graphql
  if ((env === 'lokka' || env === 'relay') && /fragment\s+on/.test(text)) {
    text = text.replace('fragment', 'fragment _');
  }

  var ast = void 0;

  try {
    ast = (0, _graphql.parse)(text);
  } catch (error) {
    context.report({
      node: node,
      message: error.message.split('\n')[0],
      loc: locFrom(node, error)
    });
    return;
  }

  var rules = env === 'relay' ? relayGraphQLValidationRules : graphQLValidationRules;

  var validationErrors = schema ? (0, _graphql.validate)(schema, ast, rules) : [];

  if (validationErrors && validationErrors.length > 0) {
    context.report({
      node: node,
      message: validationErrors[0].message,
      loc: locFrom(node, validationErrors[0])
    });
    return;
  }
}

function locFrom(node, error) {
  var location = error.locations[0];

  var line = void 0;
  var column = void 0;
  if (location.line === 1) {
    line = node.loc.start.line;
    column = node.loc.start.col + location.col;
  } else {
    line = node.loc.start.line + location.line;
    column = location.column - 1;
  }

  return {
    line: line,
    column: column
  };
}

function replaceExpressions(node, context, env) {
  var chunks = [];

  node.quasis.forEach(function (element, i) {
    var chunk = element.value.cooked;

    chunks.push(chunk);

    if (!element.tail) {
      var value = node.expressions[i];

      // Preserve location of errors by replacing with exactly the same length
      var nameLength = value.end - value.start;

      if (env === 'relay' && /:\s*$/.test(chunk)) {
        // The chunk before this one had a colon at the end, so this
        // is a variable

        // Add 2 for brackets in the interpolation
        var placeholder = strWithLen(nameLength + 2);
        chunks.push('$' + placeholder);
      } else if (env === 'lokka' && /\.\.\.\s*$/.test(chunk)) {
        // This is Lokka-style fragment interpolation where you actually type the '...' yourself
        var _placeholder = strWithLen(nameLength + 3);
        chunks.push(_placeholder);
      } else if (env === 'relay') {
        // This is Relay-style fragment interpolation where you don't type '...'
        // Ellipsis cancels out extra characters
        var _placeholder2 = strWithLen(nameLength);
        chunks.push('...' + _placeholder2);
      } else {
        // Invalid interpolation
        context.report({
          node: value,
          message: 'Invalid interpolation - not a valid fragment or variable.'
        });
        throw new Error('Invalid interpolation');
      }
    }
  });

  return chunks.join('').trim();
}

function strWithLen(len) {
  // from http://stackoverflow.com/questions/14343844/create-a-string-of-variable-length-filled-with-a-repeated-character
  return new Array(len + 1).join('x');
}

var gqlProcessor = {
  preprocess: function preprocess(text) {
    var escaped = text.replace(/`/g, '\\`');

    return [internalTag + '`' + escaped + '`'];
  },
  postprocess: function postprocess(messages) {
    // only report graphql-errors
    return (0, _lodash.flatten)(messages).filter(function (message) {
      return (0, _lodash.keys)(rules).map(function (key) {
        return 'graphql/' + key;
      }).includes(message.ruleId);
    });
  }
};

var processors = (0, _lodash.reduce)(gqlFiles, function (result, value) {
  return _extends({}, result, _defineProperty({}, '.' + value, gqlProcessor));
}, {});

module.exports = {
  rules: rules,
  processors: processors
};